

class NumpyException(Exception):
    def __init__(self, message):
        super().__init__(message)


class MatplotlibException(Exception):
    def __init__(self, message):
        super().__init__(message)


class ScipyException(Exception):
    def __init__(self, message):
        super().__init__(message)
